package eu.simstadt.nf4j;

import java.io.File;

/**
 * An export job is a proxy object for an actual nF export job. Export jobs are used to get data from your nF server. 
 * Every export job has to have a valid job descriptor and/or a job id. 
 * 
 * @author Marcel Bruse
 *
 * @param <D> The descriptor type for the export job implementation.
 */
public abstract class ExportJob<D extends ExportJobDescriptor> extends Job {

	/** Every job should have a (valid) job descriptor. */
	protected D descriptor;
	
	/**
	 * This constructor forces the job to have a description and a connector instance.
	 * 
	 * @param connector The job will use this connector to synchronize itself with the nF.
	 * @param descriptor The description of this job.
	 */
	public ExportJob(D descriptor, Connector<?, ?> connector) {
		this.descriptor = descriptor;
		this.connector = connector;
	}
	
	/**
	 * This constructor forces the job to have a id and a connector instance.
	 * 
	 * @param id The job id. If you call updateStatus() and the nF "knows" the job id, then the job status
	 * will be updated. If you call updateStatus() and the job id is "unkown" on the server side, then     
	 * @param connector The job will use this connector to synchronize itself with the nF.
	 */
	public ExportJob(int id, Connector<?, ?> connector) {
		this.id = id;
		this.connector = connector;
	}
	
	/**
	 * @return Returns the description of this job.
	 */
	public D getDescriptor() {
		return descriptor;
	}
	
	/**
	 * Once an export job has been finished, the caller should use this method to obtain the actual CityGML file.
	 * 
	 * @return A file handle to the resulting CityGML file.
	 * @throws FailedTransmissionException There could be a problem while accessing or downloading the file.
	 */
	public abstract File getResult() throws FailedTransmissionException;
	
}
