package eu.simstadt.nf4j.async;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.util.Scanner;
import java.util.zip.ZipInputStream;

import org.junit.Test;

import eu.simstadt.nf4j.FailedTransmissionException;
import eu.simstadt.nf4j.InvalidJobDescriptorException;
import eu.simstadt.nf4j.JobStatus;
import eu.simstadt.nf4j.async.AsyncExportJob;
import eu.simstadt.nf4j.async.ExportJobDescription;
import eu.simstadt.nf4j.async.HTTPConnection;
import eu.simstadt.nf4j.async.JobStatusEvent;
import eu.simstadt.nf4j.async.JobStatusListener;
import eu.simstadt.nf4j.async.Layer;
import eu.simstadt.nf4j.async.Unit;


/**
 * This class contains client oriented export job tests. It will send an export job and listens to status updates. Every
 * of the subsequent status' LOCAL, SENT, PENDING, RUNNING, FINISHED and DOWNLOAD have to be signaled to this test
 * class.
 * 
 * @author Marcel Bruse
 */
public class SuccessfulExportJob implements JobStatusListener
{

	public AsyncExportJob job;

	@Test
	public void processJob() throws InterruptedException {
		ExportJobDescription description = ExportJobDescription.getDefaultDescriptor();
		description.setInitiator(String.valueOf((int) (Math.random() * 9999)));
		String userName = System.getProperty("user.name");
		description.setAccount(userName);
		description.setProduct("WU3");
		description.setJobnumber(userName);
		description.setLODs("2");

		//FIXME: Zipped GMLs coming from nF don't have any defined srsName
		//FIXME: Save files somewhere else

		//NOTE: Unit is a predefined Map Region.
		// Some of those units are empty. 821 (Finsterrot) and 824 (Neulautern) are available on HFT nF Server.
		//		<designation>
		//      820
		//    </designation><name>
		//      Wuestenrot
		//    </name></mapsheet><mapsheet nr="127"><designation>
		//      821
		//    </designation><name>
		//      Finsterrot
		//    </name></mapsheet><mapsheet nr="128"><designation>
		//      822
		//    </designation><name>
		//      Maienfels
		//    </name></mapsheet><mapsheet nr="129"><designation>
		//      823
		//    </designation><name>
		//      Neuhütten
		//    </name></mapsheet><mapsheet nr="130"><designation>
		//      824
		//    </designation><name>
		//      Neulautern
		//    </name></mapsheet><mapsheet nr="131"><designation>
		//      824-1
		//    </designation><name>
		//      Neulautern (1)
		//    </name>
		Unit unit = Unit.getDefaultUnit();
		unit.setValue("824");
		description.addUnit(unit);


		//NOTE: Polygon selection. This would be for a small part of Neulautern
		//		ArrayList<Coord> regionPolygon = new ArrayList<>();
		//		regionPolygon.add(new Coord(49.06020829807264, 9.432239985562616));
		//		regionPolygon.add(new Coord(49.05989193639516, 9.432497477628047));
		//		regionPolygon.add(new Coord(49.05968102749148, 9.432883715726192));
		//		regionPolygon.add(new Coord(49.05935060174289, 9.433001732922845));
		//		regionPolygon.add(new Coord(49.058422585764504, 9.433066105939206));
		//		regionPolygon.add(new Coord(49.05806402949591, 9.433248496152215));
		//		regionPolygon.add(new Coord(49.05748752183746, 9.434353566266353));
		//		regionPolygon.add(new Coord(49.05788826567445, 9.435544467068967));
		//		regionPolygon.add(new Coord(49.06072150273306, 9.435233330823237));
		//		regionPolygon.add(new Coord(49.06133312328379, 9.43515822897082));
		//		regionPolygon.add(new Coord(49.06143154427858, 9.43440721044665));
		//		regionPolygon.add(new Coord(49.06020829807264, 9.432239985562616));
		//
		//		description.setRegionPolygon(regionPolygon);


		Layer layer = Layer.getDefaultLayer();
		layer.setProduct("WU3");
		layer.setName("GML");
		description.addLayer(layer);

		job = new AsyncExportJob(description, new HTTPConnection("193.196.136.164"));
		job.addJobStatusListener(this);

		try {
			job.send();
		} catch (FailedTransmissionException ex) {
			ex.printStackTrace();
		} catch (InvalidJobDescriptorException ex) {
			ex.printStackTrace();
		}

		// Wait for timeout, failure or that all tests pass
		long timeout = 1000 * 60 * 3l; // 3 minutes maximum
		long interval = 10000l;
		while (!job.hasFinished() && !job.hasFailed() && timeout > 0) {
			Thread.sleep(interval);
			timeout -= interval;
			System.out.println("+");
		}
	}

	@Override
	public void jobStatusChanged(JobStatusEvent event) {
		JobStatus status = (JobStatus) event.getSource();
		System.out.println(status);
		if (status == JobStatus.LOCAL) {
			assertTrue(true);
		} else if (status == JobStatus.SENT) {
			assertTrue(true);
		} else if (status == JobStatus.PENDING) {
			assertTrue(true);
		} else if (status == JobStatus.RUNNING) {
			assertTrue(true);
		} else if (status == JobStatus.FINISHED) {
			try {
				assertTrue(true);
				job.downloadResult();
			} catch (FailedTransmissionException ex) {
				ex.printStackTrace();
			}
		} else if (status == JobStatus.DOWNLOAD) {
			try {
				File file = job.getResult();
				assertTrue(file.canRead());
				testForExistingLOD2AndMissingLOD1(file);
			} catch (FailedTransmissionException ex) {
				ex.printStackTrace();
			}
		}
	}
	
	/**
	 * These asserts ensure that only LOD2 has been loaded as specified in the job descriptor above.
	 */
	private void testForExistingLOD2AndMissingLOD1(File file) {
		FileInputStream fileStream;
		try {
			fileStream = new FileInputStream(file);
			ZipInputStream unzipStream = new ZipInputStream(fileStream);
			unzipStream.getNextEntry(); // Skip the first entry, its the job description
			unzipStream.getNextEntry();
			File handle = Files.createTempDirectory("nfDownload").resolve("test.gml").toFile();
			BufferedOutputStream bos = 
					new BufferedOutputStream(new FileOutputStream(handle));
			byte[] in = new byte[4096];
			int read = 0;
			while ((read = unzipStream.read(in)) != -1) {
				bos.write(in, 0, read);
			}
			unzipStream.closeEntry();
			bos.close();
			unzipStream.close();
			Scanner scanner = new Scanner(handle);
			boolean lod1Missing = true;
			boolean lod2Found = false;
			while (scanner.hasNextLine()) {
				String line = scanner.nextLine();
				if (line.contains("lod2")) {
					lod2Found = true;
				}
				if (line.contains("lod1")) {
					lod1Missing = false;
				}
			}
			scanner.close();
			assertTrue(lod1Missing);
			assertTrue(lod2Found);
		} catch (IOException ex) {
			fail();
		} 
	}

}